
// Header guards
#ifndef UTILS_HEADER
#define UTILS_HEADER

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sstream>
#include <vector>
#include <bits/stdc++.h>
#include <tensor_runtime.h>
#include <tensor.h>
#include <cmath>
#include <string.h>

std::vector<float> run_accuracies;
std::string model_params_path = "../../../build/model_params/";

// FIXIT: Move this to debug.h and include in all files
void dumpWeightsToFile(const char *file_name, void *weights_ptr) {

  struct Tensor *weights = (Tensor *)weights_ptr;
  // Move data back to host
  hpvm_request_tensor(weights, 0);

  FILE *fp = fopen(file_name, "wb");
  if (fp == NULL) {
    printf("File %s could not be created. Check if directory exists \n",
           file_name);
    abort();
  }

  fclose(fp);
}

void fillTensorWithOnes(void *tensor_ptr) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  hpvm_request_tensor(tensor, 0);

  // initialization is specific to the floating point type
  if (tensor->data_type == CUDNN_DATA_FLOAT) {
    float *data_arr = (float *)tensor->host_data;
    for (unsigned int i = 0; i < tensor->num_elems; i++) {
      data_arr[i] = 1.0;
    }
  }
}

void fillWithOnesAndTwos(void *tensor_ptr) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  hpvm_request_tensor(tensor, 0);

  // initialization is specific to the floating point type
  if (tensor->data_type == CUDNN_DATA_FLOAT) {
    float *data_arr = (float *)tensor->host_data;

    for (unsigned int i = 0; i < tensor->num_elems; i++) {
      if (i % 2 == 0)
        data_arr[i] = 1.0;
      else
        data_arr[i] = 2.0;
    }

    /*for(unsigned int i = 0; i < tensor->num_elems/2; i++){
      data_arr[i] = 1.0;
    }

    for(unsigned int i = tensor->num_elems/2; i < tensor->num_elems; i++){
      data_arr[i] = 2.0;
    }*/
  }
}

void fillTensorWithVal(void *tensor_ptr, float target_value) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  hpvm_request_tensor(tensor, 0);

  // initialization is specific to the floating point type
  if (tensor->data_type == CUDNN_DATA_FLOAT) {
    float *data_arr = (float *)tensor->host_data;
    for (unsigned int i = 0; i < tensor->num_elems; i++) {
      data_arr[i] = target_value;
    }
  }
}

void fillTensorWithNegOnes(void *tensor_ptr) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  hpvm_request_tensor(tensor, 0);

  // initialization is specific to the floating point type
  if (tensor->data_type == CUDNN_DATA_FLOAT) {
    float *data_arr = (float *)tensor->host_data;
    for (unsigned int i = 0; i < tensor->num_elems; i++) {
      data_arr[i] = -1.0;
    }
  }
}

void printTensorValues(void *tensor_ptr) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  hpvm_request_tensor(tensor, 0);

  // printing is specific to the floating point type
  if (tensor->data_type != CUDNN_DATA_FLOAT) {
    // printf("\n WARNING: The tensor is non-float type tensor \n\n");
  }

  float *data_arr = (float *)tensor->host_data;

  for (unsigned int i = 0; i < tensor->num_elems; i++) {
    printf("%f,", data_arr[i]);
  }

  printf("\n");
}

void printTensorDims(void *tensor_ptr) {

  struct Tensor *tensor = (struct Tensor *)tensor_ptr;

  printf("Num_elems = %lu \n", tensor->num_elems);
  for (int i = 0; i < tensor->dims.num_dims; i++) {
    printf("dim[%d] = %lu \n", i, tensor->dims.dim_sizes[i]);
  }
}

struct Tensor *readTrainedWeights(const char *file_name, int data_type,
                                  long int dim1_size, long int dim2_size,
                                  long int dim3_size, long int dim4_size) {

  // FIXIT: Don't assume floating point types
  int type_size = 4; // NOTE: Assuming floating point tensors
  long int num_elems = dim1_size * dim2_size * dim3_size * dim4_size;
  long int size_in_bytes =
      type_size * dim1_size * dim2_size * dim3_size * dim4_size;
  float *tensor_data = (float *)malloc(sizeof(float) * num_elems);
  // printf("size_in_bytes  = %lu \n", size_in_bytes);

  int file_header_size = 0;

  FILE *file = fopen(file_name, "rb");
  if (file == NULL) {
    printf("Data file %s is not found. Aborting... \n", file_name);
    abort();
  }

  fseek(file, file_header_size, SEEK_CUR); // Skipping the file header
  fread(tensor_data, 1, size_in_bytes, file);
  fclose(file);

  struct Tensor *weights = (struct Tensor *)create4DTensor(
      data_type, nchw, dim1_size, dim2_size, dim3_size, dim4_size);

  initTensorData(weights, tensor_data, size_in_bytes);
  // compareValues(weights, tensor_data, num_elems);
  free(tensor_data);

  return weights;
}

struct Tensor *readInputBatch(const char *file_name, int data_type,
                              long int start, long int end, long int dim2_size,
                              long int dim3_size, long int dim4_size) {

  long int dim1_size = end - start;
  // FIXIT: Don't assume floating point types
  long int type_size = 4; // NOTE: Assuming floating point tensors
  long int num_elems = dim1_size * dim2_size * dim3_size * dim4_size;
  long int size_in_bytes =
      type_size * dim1_size * dim2_size * dim3_size * dim4_size;
  float *tensor_data = (float *)malloc(sizeof(float) * num_elems);
  long int file_header_size =
      type_size * start * dim2_size * dim3_size * dim4_size;

  FILE *file = fopen(file_name, "rb");
  if (file == NULL) {
    printf("Data file %s is not found. Aborting... \n", file_name);
    abort();
  }

  fseek(file, file_header_size, SEEK_SET); // Skipping the file header
  fread(tensor_data, 1, size_in_bytes, file);
  fclose(file);

  struct Tensor *weights = (struct Tensor *)create4DTensor(
      data_type, nchw, dim1_size, dim2_size, dim3_size, dim4_size);

  initTensorData(weights, tensor_data, size_in_bytes);
  free(tensor_data);

  return weights;
}

uint8_t *readLabelsBatch(const char *labels_file, int start, int end) {

  int num_labels = end - start;
  int file_header_size = sizeof(uint8_t) * start;

  uint8_t *labels = (uint8_t *)malloc(sizeof(uint8_t) * num_labels);
  FILE *file = fopen(labels_file, "rb");
  if (file == NULL) {
    printf("Data file %s is not found. Aborting...\n", labels_file);
    abort();
  }

  fseek(file, file_header_size, SEEK_SET); // Skipping the file header
  fread(labels, 1, sizeof(uint8_t) * num_labels, file);
  fclose(file);

  // printf("--labels bytes_read = %lu \n", bytes_read);
  return labels;
}

uint32_t *readLabelsBatch3(const char *labels_file, int start, int end) {

  int num_labels = end - start;
  int file_header_size = sizeof(uint32_t) * start;

  uint32_t *labels = (uint32_t *)malloc(sizeof(uint32_t) * num_labels);
  FILE *file = fopen(labels_file, "rb");
  if (file == NULL) {
    printf("Data file %s is not found. Aborting...\n", labels_file);
    abort();
  }

  fseek(file, file_header_size, SEEK_SET); // Skipping the file header
  fread(labels, 1, sizeof(uint32_t) * num_labels, file);
  fclose(file);

  return labels;
}

// NOTE: batch_size and num_classes are Unused arguments
float computeAccuracy2(uint8_t *labels, int batch_size, void *result_ptr,
                       size_t num_classes = 10) {

  struct Tensor *result = (struct Tensor *)result_ptr;

  size_t batch_dim = result->dims.dim_sizes[0];
  num_classes = result->dims.dim_sizes[1];
  float *data = (float *)result->host_data;
  int num_errors = 0;

  printf("batch_dim = %lu, channels = %lu \n", batch_dim, num_classes);

  for (unsigned int i = 0; i < batch_dim; i++) {

    int chosen = 0;
    for (size_t id = 1; id < num_classes; ++id) {
      if (data[i * num_classes + chosen] < data[i * num_classes + id])
        chosen = id;
    }

    if (chosen != labels[i])
      num_errors++;
  }

  float accuracy = ((batch_dim - num_errors) * 1.0 / batch_dim * 1.0) * 100.0;
  printf("****** Accuracy = %f \n\n", accuracy);

  FILE *fp = fopen("final_accuracy", "w+");
  if (fp != NULL) {

    std::ostringstream ss;
    ss << std::fixed << accuracy;
    std::string print_str = ss.str();

    fwrite(print_str.c_str(), 1, print_str.length(), fp);
  }

  fclose(fp);

  return accuracy;
}

float computeAccuracy3(uint32_t *labels, void *result_ptr) {

  struct Tensor *result = (struct Tensor *)result_ptr;

  size_t batch_dim = result->dims.dim_sizes[0];
  size_t num_classes = result->dims.dim_sizes[1];
  float *data = (float *)result->host_data;
  int num_errors = 0;

  printf("batch_dim = %lu, num_classes = %lu \n", batch_dim, num_classes);

  for (size_t i = 0; i < batch_dim; i++) {

    uint32_t chosen = 0;
    for (size_t id = 1; id < num_classes; ++id) {
      if (data[i * num_classes + chosen] < data[i * num_classes + id])
        chosen = id;
    }

    if (chosen != labels[i])
      num_errors++;
  }

  float accuracy = ((batch_dim - num_errors) * 1.0 / batch_dim * 1.0) * 100.0;
  printf("****** Accuracy = %f \n\n", accuracy);

  FILE *fp = fopen("final_accuracy", "w+");
  if (fp != NULL) {

    std::ostringstream ss;
    ss << std::fixed << accuracy;
    std::string print_str = ss.str();

    fwrite(print_str.c_str(), 1, print_str.length(), fp);
  }

  fclose(fp);

  return accuracy;
}

void dumpFinalAccuracy(float accuracy) {

  printf("\n\n **** Final Accuracy = %f \n", accuracy);

  FILE *fp = fopen("final_accuracy", "w+");
  if (fp != NULL) {
    std::ostringstream ss;
    ss << std::fixed << accuracy;
    std::string print_str = ss.str();

    fwrite(print_str.c_str(), 1, print_str.length(), fp);
  }

  fclose(fp);

  run_accuracies.push_back(accuracy);
}

void dumpExecutionAccuracies() {

  FILE *fp = fopen("run_accuracies.txt", "w+");
  if (fp != NULL) {
    for (size_t i = 0; i < run_accuracies.size(); i++) {
      float accuracy = run_accuracies[i];
      std::ostringstream ss;
      ss << std::fixed << accuracy;
      std::string print_str = ss.str();
      fwrite(print_str.c_str(), 1, print_str.length(), fp);
      fwrite("\n", 1, 1, fp);
    }
  }

  fclose(fp);
}
#endif
