

#include "tensor_runtime.h"
#include "utils.h"

/* NOTE: Reference Architecture to use for profiling */
void testLenetTanh() {
  int total_runs = 1;
  printf("********* Lenet-2 Architecture ********** \n");
  // FIXIT: Extend this to batch of images - currently 5 images

  int test_batch_size = 5000;

  std::string dir_prefix = model_params_path + std::string("/lenet_mnist/");

  std::string input_path = dir_prefix + std::string("input.bin");
  std::string labels_path = dir_prefix + std::string("labels.bin");
  std::string labels32_path = dir_prefix + std::string("labels32.bin");

  // Loading Input Batch
  void *input =
      readInputBatch(input_path.c_str(), 0, 0, test_batch_size, 1, 28, 28);
  uint8_t *labels = readLabelsBatch(labels_path.c_str(), 0, test_batch_size);

  void *conv1_filter = readTrainedWeights(
      "../model_params/lenet_mnist/conv1.bin", float_type, 32, 1, 5, 5);
  void *conv1_bias = readTrainedWeights(
      "../model_params/lenet_mnist/conv1_bias.bin", float_type, 1, 32, 1, 1);
  void *conv2_filter = readTrainedWeights(
      "../model_params/lenet_mnist/conv2.bin", float_type, 64, 32, 5, 5);
  void *conv2_bias = readTrainedWeights(
      "../model_params/lenet_mnist/conv2_bias.bin", float_type, 1, 64, 1, 1);
  void *fc1_weights = readTrainedWeights("../model_params/lenet_mnist/fc1.bin",
                                         float_type, 1, 1, 7 * 7 * 64, 1024);
  void *fc1_bias = readTrainedWeights(
      "../model_params/lenet_mnist/fc1_bias.bin", float_type, 1, 1024, 1, 1);
  void *fc2_weights = readTrainedWeights("../model_params/lenet_mnist/fc2.bin",
                                         float_type, 1, 1, 1024, 10);
  void *fc2_bias = readTrainedWeights(
      "../model_params/lenet_mnist/fc2_bias.bin", float_type, 1, 10, 1, 1);

  clearTensorMap();

  for (int i = 0; i < total_runs; i++) {
    readOpenTunerFlags("opentuner_flags"); // Resets the OpenTuner counters

    // Start power and performnce profiling
    startProfiling();

    int conv_mode = 1; // NOTE: using CROSS_CORRELATION
    int conv_precision =
        0; // NOTE: using Float as compute precision. FIXIT: use enum

    // NOTE: 'SAME' convolution
    void *conv1out = tensorHalfConvolution(input, conv1_filter, 2, 2, 1, 1,
                                           conv_mode, conv_precision);

    // NOTE: For tensorAdd, the only dimension that MUST match is channels
    tensorHalfAdd(conv1out, conv1_bias); // NOTE: In place operation

    void *pool1out = tensorHalfPooling(conv1out, 0, 2, 2, 0, 0, 2, 2);

    void *conv1_tanh = tensorHalfTanh(pool1out);

    // NOTE: input channels have to match between tensor op inputs and outputs
    void *conv2out = tensorHalfConvolution(conv1_tanh, conv2_filter, 2, 2, 1, 1,
                                           conv_mode, conv_precision);
    tensorHalfAdd(conv2out, conv2_bias); // NOTE: In place operation

    void *pool2out = tensorHalfPooling(conv2out, 0, 2, 2, 0, 0, 2, 2);

    void *conv2_tanh = tensorHalfTanh(pool2out);

    void *gemm1out = tensorHalfGemm(conv2_tanh, fc1_weights);

    void *gemm1biasout = tensorHalfAdd(gemm1out, fc1_bias);

    void *tanh1out = tensorHalfTanh(gemm1biasout);

    void *gemm2out = tensorHalfGemm(tanh1out, fc2_weights);

    void *gemm2_biasout = tensorHalfAdd(gemm2out, fc2_bias);

    void *tanh2out = tensorHalfTanh(gemm2_biasout);

    void *result = tensorSoftmax(tanh2out);

    // End profiling and dump output to profile.txt
    stopProfiling();

    computeAccuracy2(labels, test_batch_size, result);

    dumpAccuracyNorms();
    freeOutputTensors();
  }
}

int main(int argc, char *argv[]) {
  llvm_hpvm_initTensorRt(0);

  testLenetTanh();

  llvm_hpvm_cleanupTensorRt();

  return 0;
}
